from dataclasses import dataclass as _dataclass
import abc as _abc

from trading_sdk.core import SDK
from .data import MarketData, PerpMarketData
from .user import UserData, PerpUserData
from .trade import Trading

@_dataclass(frozen=True)
class Market(SDK):
  Data = MarketData
  User = UserData
  Trade = Trading
  data: MarketData
  user: UserData
  trade: Trading

  @property
  @_abc.abstractmethod
  def venue(self) -> str:
    """Venue identifier/name."""

  @property
  @_abc.abstractmethod
  def market_id(self) -> str:
    """Market identifier/name."""
    ...

  @property
  def id(self) -> str:
    return f'{self.venue}:{self.market_id}'

@_dataclass(frozen=True)
class PerpMarket(Market):
  Data = PerpMarketData
  User = PerpUserData
  data: PerpMarketData
  user: PerpUserData


from typing_extensions import AsyncIterable, Sequence
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime

from trading_sdk.core import SDK

@dataclass
class FundingRate:
  rate: Decimal
  """Funding rate (in relative units, e.g. 0.01 = 1%)."""
  time: datetime
  """Funding payment time."""

class Funding(SDK):
  Funding = FundingRate
  
  @SDK.method
  @abstractmethod
  def history(
    self, start: datetime, end: datetime,
  ) -> AsyncIterable[Sequence[Funding]]:
    """Fetch perpetual funding rate history."""

  async def history_sync(self, start: datetime, end: datetime) -> Sequence[Funding]:
    """Fetch perpetual funding rate history without streaming."""
    out: list[Funding.Funding] = []
    async for page in self.history(start, end):
      out.extend(page)
    return out

  @SDK.method
  @abstractmethod
  async def next(self) -> Funding:
    """Fetch the next funding rate and time."""



from dataclasses import dataclass as _dataclass

from trading_sdk.core import SDK
from .place import Place
from .cancel import Cancel

@_dataclass(frozen=True)
class Trading(SDK):
	Place = Place
	Cancel = Cancel
	place: Place
	cancel: Cancel


from typing_extensions import AsyncIterable, TypeVar, ParamSpec, Generic, Sequence, Callable, Awaitable
from decimal import Decimal, ROUND_HALF_DOWN, ROUND_FLOOR, ROUND_CEILING

T = TypeVar('T')
U = TypeVar('U')
P = ParamSpec('P')


def round2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).quantize(Decimal('1.'), rounding=ROUND_HALF_DOWN) * tick_size
  return r.normalize()


def trunc2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).to_integral_value(rounding=ROUND_FLOOR) * tick_size
  return r.normalize()

def ceil2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).to_integral_value(rounding=ROUND_CEILING) * tick_size
  return r.normalize()

class Stream(AsyncIterable[T], Awaitable[Sequence[T]], Generic[T]):
  def __init__(self, xs: AsyncIterable[T]):
    self.xs = xs

  def __aiter__(self):
    return self.xs.__aiter__()

  def __await__(self):
    async def coro():
      return [x async for x in self]
    return coro().__await__()

  @staticmethod
  def lift(fn: Callable[P, AsyncIterable[U]]):
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Stream[U]:
      return Stream(fn(*args, **kwargs))
    return wrapper
  

class ChunkedStream(AsyncIterable[Sequence[T]], Awaitable[Sequence[T]], Generic[T]):
  def __init__(self, xs: AsyncIterable[Sequence[T]]):
    self.xs = xs

  def __aiter__(self):
    return self.xs.__aiter__()

  def __await__(self):
    async def coro():
      return [x async for xs in self for x in xs]
    return coro().__await__()

  async def flatten(self) -> AsyncIterable[T]:
    async for xs in self:
      for x in xs:
        yield x

  @staticmethod
  def lift(fn: Callable[P, AsyncIterable[Sequence[U]]]):
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> ChunkedStream[U]:
      return ChunkedStream(fn(*args, **kwargs))
    return wrapper


from typing_extensions import AsyncIterable, Sequence
from abc import abstractmethod
from datetime import datetime

from trading_sdk.core import ChunkedStream, SDK
from .types import Transaction

class Transactions(SDK):
  def transactions(
    self, start: datetime, end: datetime
  ) -> ChunkedStream[Transaction]:
    """Collect all transactions of the account, in a given time range."""
    return ChunkedStream(self._transactions_impl(start=start, end=end))
  
  @SDK.method
  @abstractmethod
  def _transactions_impl(
    self, start: datetime, end: datetime
  ) -> AsyncIterable[Sequence[Transaction]]:
    ...


from .exc import Error, NetworkError, ValidationError, UserError, AuthError, ApiError, LogicError
from .misc import Num, fmt_num
from .sdk import SDK, instrument, exponential_retry, log
from .util import round2tick, trunc2tick, ceil2tick, Stream, ChunkedStream


from abc import abstractmethod
from decimal import Decimal
from dataclasses import dataclass

from trading_sdk.core import SDK
	
class Position(SDK):
  @dataclass
  class Position:
    size: Decimal = Decimal(0)
    """Position size (of base units)"""

  @SDK.method
  @abstractmethod
  async def get(self) -> Position:
    """Fetch your (base) balance/position in the market."""

  @SDK.method
  async def __call__(self) -> Position:
    """Fetch your (base) balance/position in the market."""
    return await self.get()


class PerpPosition(Position):
  @dataclass(kw_only=True)
  class Position(Position.Position):
    entry_price: Decimal = Decimal(0)
    """Position average entry price (in quote units). Only meaningful if size != 0"""
    
  @SDK.method
  @abstractmethod
  async def get(self) -> Position:
    """Fetch your (base) position in the market."""
  
  @SDK.method
  async def __call__(self) -> Position:
    """Fetch your (base) position in the market."""
    return await self.get()


from typing_extensions import Literal, Sequence
from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from trading_sdk import SDK

@dataclass(kw_only=True)
class Snapshot:
  asset: str
  time: datetime
  qty: Decimal
  kind: Literal['currency', 'future', 'strategy']
  avg_price: Decimal | None = None
  """Average entry price"""

class Snapshots(SDK):
  @SDK.method
  @abstractmethod
  async def snapshots(self, assets: Sequence[str] = []) -> Sequence[Snapshot]:
    """Snapshot the portfolio of the account.
    
    - `assets`: hint of assets to snapshot. Behavior depends on the implementation.
      - For platforms that can snapshot all assets, this is ignored.
      - For platforms where all assets can't be known a priori (e.g. a blockchain), only the given `assets` are queried.
    """
    ...


from typing_extensions import Literal, TypeVar, Any, Sequence, Union, Generic, Annotated
from dataclasses import dataclass, replace, field
from decimal import Decimal
from datetime import datetime

D = TypeVar('D', default=Any, covariant=True)
D2 = TypeVar('D2', default=Any, covariant=True)

Label = Literal[
  'trade', 'yield', 'fee', 'bonus', 'other', 'internal_transfer',
  'future_trade', 'settlement', 'settlement_fee', 'funding',
  'strategy_deposit', 'strategy_withdrawal',
  'crypto_deposit', 'fiat_deposit', 'crypto_withdrawal', 'fiat_withdrawal', 'withdrawal_fee',
  'ethereum_transfer', 'erc20_transfer',
]

@dataclass(kw_only=True, frozen=True)
class Flow(Generic[D]):
  Label = Label
  kind: Literal['currency', 'future', 'strategy'] | None = None
  asset: str
  change: Decimal
  label: Label
  time: datetime
  price: Decimal | None = None
  details: D

@dataclass(kw_only=True, frozen=True)
class BaseEvent(Generic[D]):
  id: str
  time: datetime
  details: D

  @property
  def expected_flows(self) -> Sequence[Flow[D]]:
    """Flows expected to be matched."""
    return []

  @property
  def fixed_flows(self) -> Sequence[Flow[D]]:
    """Flows not appearing in the flows list but that should be included in the event."""
    return []

  @property
  def flows(self) -> Sequence[Flow[D]]:
    return list(self.expected_flows) + list(self.fixed_flows)


@dataclass(frozen=True)
class Fee:
  amount: Decimal
  asset: str

def sign(side: Literal['buy', 'sell']) -> Literal[1, -1]:
  match side:
    case 'buy': return 1
    case 'sell': return -1
    case _: raise ValueError(f'Unknown side: {side}')

@dataclass(kw_only=True, frozen=True)
class Trade(BaseEvent[D], Generic[D]):
  type: Literal['trade'] = field(default='trade')
  base: str
  quote: str
  qty: Decimal
  """Base asset quantity."""
  price: Decimal
  liquidity: Literal['maker', 'taker']
  side: Literal['buy', 'sell']
  fee: Fee | None = None

  @property
  def expected_flows(self) -> list[Flow[D]]:
    s = sign(self.side)
    flows = [
      Flow(
        time=self.time,
        label='trade',
        asset=self.base,
        change=s * self.qty,
        details=self.details,
        kind='currency',
      ),
      Flow(
        time=self.time,
        label='trade',
        asset=self.quote,
        change=-s * self.qty * self.price,
        details=self.details,
        kind='currency',
      ),
    ]
    if self.fee is not None:
      flows.append(Flow(
        time=self.time,
        label='fee',
        asset=self.fee.asset,
        change=-self.fee.amount,
        details=self.details,
        kind='currency',
      ))
    return flows
    

@dataclass(kw_only=True, frozen=True)
class FutureTrade(BaseEvent[D], Generic[D]):
  type: Literal['future_trade'] = field(default='future_trade')
  instrument: str
  size: Decimal
  price: Decimal
  liquidity: Literal['maker', 'taker']
  side: Literal['buy', 'sell']
  fee: Fee | None = None

  @property
  def expected_flows(self) -> list[Flow[D]]:
    if self.fee is not None:
      return [
        Flow(
          time=self.time,
          label='fee',
          asset=self.fee.asset,
          change=-self.fee.amount,
          details=self.details,
          kind='currency',
        ),
      ]
    else:
      return []

  @property  
  def fixed_flows(self) -> list[Flow[D]]:
    s = sign(self.side)
    return [
      Flow(
        time=self.time,
        label='future_trade',
        asset=self.instrument,
        change=s*self.size,
        price=self.price,
        details=self.details,
        kind='future',
      )
    ]

@dataclass(kw_only=True, frozen=True)
class SingleEvent(BaseEvent[D], Generic[D]):
  asset: str
  qty: Decimal
  type: Flow.Label

  @property
  def expected_flows(self) -> list[Flow[D]]:
    return [
      Flow(
        time=self.time,
        label=self.type,
        asset=self.asset,
        change=self.qty,
        details=self.details,
        kind='currency',
      )
    ]

  @classmethod
  def of(cls, id: str, flow: Flow[D2]) -> 'Event[D2]':
    match flow.label:
      case 'yield': cls = Yield
      case 'bonus': cls = Bonus
      case 'funding': cls = Funding
      case 'settlement': cls = Settlement
      case 'fee': cls = FeeEvent
      case 'internal_transfer': cls = InternalTransfer
      case 'other': cls = Other
      case 'strategy_deposit' | 'strategy_withdrawal':
        return Strategy(type=flow.label, id=id, time=flow.time, asset=flow.asset, qty=flow.change, details=flow.details)
      case _: raise ValueError(f'Unknown operation type: {flow.label}')
    return cls(id=id, time=flow.time, asset=flow.asset, qty=flow.change, details=flow.details)


@dataclass(kw_only=True, frozen=True)
class FeeEvent(SingleEvent[D], Generic[D]):
  type: Literal['fee'] = field(default='fee')

@dataclass(kw_only=True, frozen=True)
class InternalTransfer(SingleEvent[D], Generic[D]):
  type: Literal['internal_transfer'] = field(default='internal_transfer')

@dataclass(kw_only=True, frozen=True)
class Strategy(SingleEvent[D], Generic[D]):
  """Deposit into or withdraw from a strategy (E.g. trading bot, copy trading, etc.)"""
  strategy: str | None = None
  type: Literal['strategy_deposit', 'strategy_withdrawal']

  @property
  def fixed_flows(self) -> list[Flow[D]]:
    return [
      Flow(
        time=self.time,
        label=self.type,
        asset=self.strategy or self.asset,
        kind='strategy',
        change=-self.qty,
        details=self.details,
      )
    ]

@dataclass(kw_only=True, frozen=True)
class Yield(SingleEvent[D], Generic[D]):
  type: Literal['yield'] = field(default='yield')

@dataclass(kw_only=True, frozen=True)
class Other(SingleEvent[D], Generic[D]):
  type: Literal['other'] = field(default='other')

@dataclass(kw_only=True, frozen=True)
class Bonus(SingleEvent[D], Generic[D]):
  type: Literal['bonus'] = field(default='bonus')

@dataclass(kw_only=True, frozen=True)
class Funding(SingleEvent[D], Generic[D]):
  type: Literal['funding'] = field(default='funding')

@dataclass(kw_only=True, frozen=True)
class Settlement(SingleEvent[D], Generic[D]):
  type: Literal['settlement'] = field(default='settlement')

@dataclass(kw_only=True, frozen=True)
class CryptoDeposit(SingleEvent[D], Generic[D]):
  """Deposit from a blockchain."""
  type: Literal['crypto_deposit'] = field(default='crypto_deposit')
  network: str
  tx_hash: str
  idx: int | None = None
  """Index of the event within the same transaction."""

@dataclass(kw_only=True, frozen=True)
class FiatDeposit(SingleEvent[D], Generic[D]):
  type: Literal['fiat_deposit'] = field(default='fiat_deposit')
  method: str
  fiat_currency: str
  """Fiat currency used to buy the asset."""
  fiat_amount: Decimal
  """Amount of the fiat currency used to buy the asset."""

@dataclass(kw_only=True, frozen=True)
class Withdrawal(SingleEvent[D], Generic[D]):
  fee: Fee | None = None

  @property
  def expected_flows(self) -> list[Flow[D]]:
    flows = [
      Flow(
        time=self.time,
        label=self.type,
        asset=self.asset,
        change=-self.qty,
        details=self.details,
        kind='currency',
      )
    ]
    if self.fee is not None:
      flows.append(Flow(
        time=self.time,
        label='withdrawal_fee',
        asset=self.fee.asset,
        change=-self.fee.amount,
        details=self.details,
        kind='currency',
      ))
    return flows

@dataclass(kw_only=True, frozen=True)
class CryptoWithdrawal(Withdrawal[D], Generic[D]):
  """Withdrawal into a blockchain."""
  type: Literal['crypto_withdrawal'] = field(default='crypto_withdrawal')
  network: str
  tx_hash: str
  idx: int | None = None
  """Index of the event within the same transaction."""
  address: str
  memo: str | None = None

@dataclass(kw_only=True, frozen=True)
class FiatWithdrawal(Withdrawal[D], Generic[D]):
  type: Literal['fiat_withdrawal'] = field(default='fiat_withdrawal')
  method: str
  fiat_currency: str
  """Fiat currency received."""
  fiat_amount: Decimal
  """Amount of the fiat currency received."""


@dataclass(kw_only=True, frozen=True)
class BaseEthereumTransaction(BaseEvent[D], Generic[D]):
  tx_hash: str
  from_address: str
  to_address: str
  chain_id: int
  """Chain ID (see https://chainid.network/)"""

@dataclass(kw_only=True, frozen=True)
class EthereumTransaction(BaseEthereumTransaction[D], Generic[D]):
  type: Literal['ethereum_transfer'] = field(default='ethereum_transfer')
  value: Decimal
  """Value sent [ETH]"""
  fee: Decimal | None = None
  """Fee paid [ETH]"""

  @property
  def expected_flows(self) -> list[Flow[D]]:
    flows: list[Flow[D]] = []
    if self.value:
      flows.append(Flow(
        time=self.time,
        label='ethereum_transfer',
        asset='ETH',
        change=self.value,
        details=self.details,
        kind='currency',
      ))
    if self.fee is not None:
      flows.append(Flow(
        time=self.time,
        label='fee',  
        asset='ETH',
        change=-self.fee,
        details=self.details,
        kind='currency',
      ))
    return flows

@dataclass(kw_only=True, frozen=True)
class ERC20Transfer(BaseEthereumTransaction[D], Generic[D]):
  @dataclass
  class Transfer:
    sender_address: str
    recipient_address: str
    contract_address: str
    value: Decimal
    """Value sent [Token]"""
    direction: Literal['IN', 'OUT']

  type: Literal['erc20_transfer'] = field(default='erc20_transfer')
  fee: Decimal | None = None
  """Fee paid [ETH]"""
  transfers: Sequence[Transfer]

  @property
  def expected_flows(self) -> list[Flow[D]]:
    flows = [
      Flow(
        time=self.time,
        label='erc20_transfer',
        asset=transfer.contract_address,
        change=transfer.value * (1 if transfer.direction == 'IN' else -1),
        details=self.details,
        kind='currency',
      )
      for transfer in self.transfers
    ]
    if self.fee:
      flows.append(Flow(
        time=self.time,
        label='fee',
        asset='ETH',
        change=-self.fee,
        details=self.details,
        kind='currency',
      ))
    return flows

Event = Union[
  Trade[D], FutureTrade[D], Strategy[D], InternalTransfer[D],
  Yield[D], Bonus[D], Funding[D], Settlement[D], Other[D], FeeEvent[D],
  CryptoDeposit[D], FiatDeposit[D], CryptoWithdrawal[D], FiatWithdrawal[D],
  EthereumTransaction[D], ERC20Transfer[D],
]

@dataclass(kw_only=True, frozen=True)
class Transaction(Generic[D]):
  event: Event[D]
  flows: Sequence[Flow[D]]

  @property
  def time(self):
    return self.event.time

  def replace_time(self, time: datetime) -> 'Transaction[D]':
    return replace(
      self, event=replace(self.event, time=time),
      flows=[replace(p, time=time) for p in self.flows]
    )

  @staticmethod
  def single(id: str, flow: Flow[D2]) -> 'Transaction[D2]':
    event = SingleEvent[D2].of(id, flow)
    return Transaction(event=event, flows=[flow] + list(event.fixed_flows))


from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal

from trading_sdk.core import SDK

  
class Balances(SDK):
  @dataclass
  class Balance:
    free: Decimal = Decimal(0)
    locked: Decimal = Decimal(0)

    @property
    def total(self) -> Decimal:
      return self.free + self.locked
      
  @SDK.method
  @abstractmethod
  async def quote(self) -> Balance:
    """Fetch the quote/collateral currency balance."""


from dataclasses import dataclass as _dataclass

from trading_sdk.core import SDK
from .balances import Balances
from .position import Position, PerpPosition
from .trades import Trades
from .orders import Orders
from .funding import Funding

@_dataclass(frozen=True)
class UserData(SDK):
	Balances = Balances
	Position = Position
	Trades = Trades
	Orders = Orders
	balances: Balances
	position: Position
	trades: Trades
	orders: Orders
	
@_dataclass(frozen=True)
class PerpUserData(UserData):
	Position = PerpPosition
	Funding = Funding
	position: PerpPosition
	funding: Funding


from typing_extensions import Sequence, AsyncIterable, Any
from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from trading_sdk.core import SDK

class Trades(SDK):

  @dataclass(kw_only=True)
  class Trade:
    @dataclass(kw_only=True)
    class Fee:
      amount: Decimal
      """Fee paid (or received if negative, in fee asset units)."""
      asset: str

    id: str | None
    price: Decimal
    qty: Decimal
    """Signed quantity (netagive -> sell, positive -> buy)"""
    time: datetime
    maker: bool
    fee: Fee | None = None
    details: Any = None

  @SDK.method
  @abstractmethod
  def history(self, start: datetime, end: datetime) -> AsyncIterable[Sequence[Trade]]:
    """Fetch your trades."""

  async def history_sync(self, start: datetime, end: datetime) -> Sequence[Trade]:
    """Fetch your trades without streaming."""
    out: list[Trades.Trade] = []
    async for page in self.history(start, end):
      out.extend(page)
    return out
		
  @SDK.method
  @abstractmethod
  def stream(self) -> AsyncIterable[Trade]:
    """Subscribe to your real-time trades."""


from .core import (
  Num, fmt_num, SDK,
  Error, NetworkError, ValidationError, UserError, AuthError, ApiError, LogicError,
)
from .earn import Earn
from .market import Market, PerpMarket
from .wallet import Wallet

__all__ = [
  'SDK',
  'Earn', 'Wallet', 'Market', 'PerpMarket',
  'Num', 'fmt_num',
  'Error', 'NetworkError', 'ValidationError', 'UserError', 'AuthError', 'ApiError', 'LogicError',
]


from dataclasses import dataclass as _dataclass
from trading_sdk.core import SDK

from .rules import Rules
from .depth import Depth
from .funding import Funding
from .index import Index

@_dataclass(frozen=True)
class MarketData(SDK):
	Rules = Rules
	Depth = Depth
	rules: Rules
	depth: Depth

@_dataclass(frozen=True)
class PerpMarketData(MarketData):
	Funding = Funding
	Index = Index
	funding: Funding
	index: Index


from typing_extensions import TypedDict, Literal, Sequence, Any
from abc import abstractmethod
from dataclasses import dataclass
import asyncio

from trading_sdk.core import Num, SDK


class Place(SDK):
  class Order(TypedDict):
    qty: Num
    """Quantity of the order in the base asset. Negative -> sell, positive -> buy."""
    price: Num
    type: Literal['LIMIT', 'POST_ONLY']
    """No, market orders are not supported. Purposefully."""


  @dataclass(kw_only=True)
  class Result:
    id: str
    details: Any = None
  
  @SDK.method
  @abstractmethod
  async def order(self, order: Order) -> Result:
    """Place an order in the market."""
  
  @SDK.method
  async def orders(self, orders: Sequence[Order]) -> Sequence[Result]:
    """Place multiple orders in the market."""
    return await asyncio.gather(*[self.order(order) for order in orders])


from .types import (
  Transaction, Event, Flow, Label,
  Fee, Trade, FutureTrade, SingleEvent, Strategy, InternalTransfer,
  Yield, Bonus, Funding, Settlement, Other,
  CryptoDeposit, FiatDeposit, CryptoWithdrawal, FiatWithdrawal,
  EthereumTransaction, ERC20Transfer,
)
from .sdk import Transactions
from .matching import match_transactions, EventMatcher


from typing_extensions import Literal, Collection, Sequence, ClassVar
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from datetime import timedelta

from trading_sdk.core import SDK

InstrumentTag = Literal['flexible', 'fixed', 'one-time', 'new-users', 'staking']

@dataclass(kw_only=True)
class Instrument:
	Tag: ClassVar = InstrumentTag
	tags: Sequence[InstrumentTag]
	asset: str
	apr: Decimal
	"""Annual Percent Rate, as a fraction of 1 (0.01 = 1%)"""
	yield_asset: str | None = None
	"""Asset that yields the interest. If not provided, it's the same as the subscribed asset."""
	min_qty: Decimal | None = None
	"""Minimum quantity of the asset to invest"""
	max_qty: Decimal | None = None
	"""Maximum quantity of the asset to invest"""
	url: str | None = None
	duration: timedelta | None = None
	"""Duration of the instrument"""
	id: str | None = None
	"""Unique identifier for the instrument, if available."""

class Instruments(SDK):
  @SDK.method
  @abstractmethod
  async def instruments(
    self, *, tags: Collection[InstrumentTag] | None = None,
    assets: Collection[str] | None = None,
  ) -> Sequence[Instrument]:
    """Fetch instruments from the exchange.

    - `tags`: Filter by tags. Returns all assets matching at least one of the tags.
    - `assets`: Filter by (subscription) assets.
    """


from typing_extensions import Sequence
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal

from trading_sdk.core import SDK

@dataclass(kw_only=True)
class DepositMethod:
  @dataclass
  class Fee:
    asset: str
    amount: Decimal

  asset: str
  network: str
  fee: Fee | None = None
  contract_address: str | None = None
  min_confirmations: int | None = None

class DepositMethods(SDK):
  @SDK.method
  @abstractmethod
  async def deposit_methods(
    self, *, assets: Sequence[str] | None = None,
  ) -> Sequence[DepositMethod]:
    """Get deposit methods.

    - `assets`: optional filter by asset.
    """


from abc import abstractmethod
from decimal import Decimal

from trading_sdk.core import SDK

class Index(SDK):
	@SDK.method
	@abstractmethod
	async def price(self) -> Decimal:
		"""Fetch the market's index price."""

	@SDK.method
	async def __call__(self) -> Decimal:
		"""Fetch the market's index price."""
		return await self.price()


from typing_extensions import Sequence, AsyncIterable
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime

from trading_sdk.core import SDK

  
class Funding(SDK):
  @dataclass
  class Payment:
    amount: Decimal
    """Funding paid (if positive) or received (in quote units)"""
    time: datetime

  @SDK.method
  @abstractmethod
  def history(self, start: datetime, end: datetime) -> AsyncIterable[Sequence[Payment]]:
    """Fetch your funding payments."""

  async def history_sync(self, start: datetime, end: datetime) -> Sequence[Payment]:
    """Fetch your funding payments without streaming."""
    out: list[Funding.Payment] = []
    async for page in self.history(start, end):
      out.extend(page)
    return out


from typing_extensions import Sequence, overload
from abc import abstractmethod
from dataclasses import dataclass, field, replace
from decimal import Decimal

from trading_sdk.core import SDK

@dataclass(kw_only=True)
class Book:
  
  @dataclass
  class Entry:
    price: Decimal
    qty: Decimal
    """Quantity in base units."""

    @property
    def notional(self) -> Decimal:
      return self.price * self.qty

    def __str__(self) -> str:
      return f'{self:f}'

    def __format__(self, fmt: str) -> str:
      return f'[{self.price:{fmt}}] {self.qty:{fmt}} (${self.notional:{fmt}})'

  bids: list[Entry] = field(default_factory=list)
  """Bids, sorted by price descending (best bid first)."""
  asks: list[Entry] = field(default_factory=list)
  """Asks, sorted by price ascending (best ask first)."""

  def __post_init__(self):
    self.bids.sort(key=lambda e: e.price, reverse=True)
    self.asks.sort(key=lambda e: e.price)

  @property
  def best_bid(self) -> Entry:
    return max(self.bids, key=lambda e: e.price)

  @property
  def best_ask(self) -> Entry:
    return min(self.asks, key=lambda e: e.price)

  @overload
  def market_buy_price(self, *, qty: Decimal) -> Decimal | None:
    """Average fill price for a market order buying `qty` base units."""
  @overload
  def market_buy_price(self, *, notional: Decimal) -> Decimal | None:
    """Average fill price for a market order buying `notional` value."""
  def market_buy_price(self, *, qty: Decimal | None = None, notional: Decimal | None = None) -> Decimal | None:
    if qty is not None:
      return market_price_qty(self.asks, qty)
    elif notional is not None:
      return market_price_notional(self.asks, notional)
    else:
      raise ValueError("Either qty or notional must be provided")

  @overload
  def market_sell_price(self, *, qty: Decimal) -> Decimal | None:
    """Average fill price for a market order selling `qty` base units."""
  @overload
  def market_sell_price(self, *, notional: Decimal) -> Decimal | None:
    """Average fill price for a market order selling `notional` value."""
  def market_sell_price(self, *, qty: Decimal | None = None, notional: Decimal | None = None) -> Decimal | None:
    if qty is not None:
      return market_price_qty(self.bids, qty)
    elif notional is not None:
      return market_price_notional(self.bids, notional)
    else:
      raise ValueError("Either qty or notional must be provided")

  def sellable_at(self, price: Decimal) -> Decimal:
    """Max. sellable quantity that will fill at average price >= `price`."""
    return sellable_qty(self.bids, price)

  def buyable_at(self, price: Decimal) -> Decimal:
    """Max. buyable quantity that will fill at average price <= `price`."""
    return buyable_qty(self.asks, price)

  def buy(self, *, qty: Decimal) -> Decimal | None:
    """Buy `qty` base units at the best price, returning the average fill price.
    
    #### Warning: Mutates the book in place.
    """
    return fill(self.asks, qty=qty)

  def sell(self, *, qty: Decimal) -> Decimal | None:
    """Sell `qty` base units at the best price, returning the average fill price.
    
    #### Warning: Mutates the book in place.
    """
    return fill(self.bids, qty=qty)

  def merge(self, *others: 'Book') -> 'Book':
    return Book(
      bids=self.bids + [e for other in others for e in other.bids],
      asks=self.asks + [e for other in others for e in other.asks],
    )

  def with_fees(self, fee: Decimal) -> 'Book':
    """Lower bids and raise asks to account for fees."""
    return Book(
      bids=[Book.Entry(e.price * (1 - fee), e.qty) for e in self.bids],
      asks=[Book.Entry(e.price * (1 + fee), e.qty) for e in self.asks],
    )

  def limit(self, levels: int) -> 'Book':
    """Limit the book to `levels` levels."""
    return Book(
      bids=self.bids[:levels],
      asks=self.asks[:levels],
    )

  @property
  def mark_price(self) -> Decimal:
    return (self.best_bid.price + self.best_ask.price) / 2

  def __str__(self) -> str:
    return f'{self:f}'

  def __format__(self, fmt: str) -> str:
    return fmt_book(self, fmt)


class Depth(SDK):
  Book = Book

  @SDK.method
  @abstractmethod
  async def book(self, *, limit: int | None = None) -> Book:
    """Fetch the market order book.
    
    - `limit`: The maximum number of bids/asks to return.
    """

  @SDK.method
  async def __call__(self, *, limit: int | None = None) -> Book:
    """Fetch the market order book.
    
    - `limit`: The maximum number of bids/asks to return.
    """
    return await self.book(limit=limit)



def avg_price(entries: Sequence['Book.Entry']) -> Decimal:
  total = sum(e.price * e.qty for e in entries)
  total_qty = sum(e.qty for e in entries)
  return Decimal(total) / Decimal(total_qty)


def market_price_qty(entries: list[Book.Entry], qty: Decimal) -> Decimal | None:
  """Average fill price for a market order consuming `qty` base units."""
  remaining = qty
  trades: list[Book.Entry] = []
  i = 0
  while remaining > 0 and i < len(entries):
    entry = entries[i]
    if entry.qty <= remaining:
      remaining -= entry.qty
      trades.append(entry)
      i += 1
    else:
      trades.append(Book.Entry(entry.price, remaining))
      remaining = Decimal(0)

  if remaining == 0:
    return avg_price(trades)


def market_price_notional(entries: list[Book.Entry], notional: Decimal) -> Decimal | None:
  """Average fill price for a market order consuming `notional` value."""
  remaining = notional
  trades: list[Book.Entry] = []
  i = 0
  while remaining > 0 and i < len(entries):
    entry = entries[i]
    if entry.notional <= remaining:
      remaining -= entry.notional
      trades.append(entry)
      i += 1
    else:
      trades.append(Book.Entry(entry.price, remaining / entry.price))
      remaining = Decimal(0)
  
  if remaining == 0:
    return avg_price(trades)


def buyable_qty(asks: list[Book.Entry], price: Decimal) -> Decimal:
  """Max. fillable quantity with average fill price <= `price`."""
  Q = Decimal(0) # total qty
  N = Decimal(0) # total notional
  P = price # target price
  for e in asks:
    if e.price <= P:
      Q += e.qty
      N += e.qty*e.price
    else:
      q = min((N - P*Q) / (P - e.price), e.qty)
      Q += q
      N += q*e.price
      if q < e.qty:
        break
  return Q


def sellable_qty(bids: list[Book.Entry], price: Decimal) -> Decimal:
  """Max. fillable quantity with average fill price >= `price`."""
  Q = Decimal(0) # total qty
  N = Decimal(0) # total notional
  P = price # target price
  for e in bids:
    if e.price >= P:
      Q += e.qty
      N += e.qty*e.price
    else:
      q = min((N - P*Q) / (P - e.price), e.qty)
      Q += q
      N += q*e.price
      if q < e.qty:
        break
  return Q

def fill(entries: list[Book.Entry], *, qty: Decimal) -> Decimal | None:
  """Fill `qty` base units of the book, returning the average fill price.
  
  #### Warning: Mutates the book in place.
  """
  orig_qty = qty
  notional = Decimal(0)

  while qty > 0 and entries:
    e = entries[0]
    if e.qty <= qty:
      entries.pop(0)
      notional += e.qty*e.price
      qty -= e.qty
    else:
      e.qty -= qty
      notional += qty*e.price
      qty = Decimal(0)

  if qty == 0:
    return notional / orig_qty

def with_fees(self: 'Book', fee: Decimal) -> 'Book':
  """Lower bids and raise asks to account for fees."""
  return Book(
    bids=[Book.Entry(e.price * (1-fee), e.qty) for e in self.bids],
    asks=[Book.Entry(e.price * (1+fee), e.qty) for e in self.asks],
  )

def fmt_book(self: 'Book', fmt: str) -> str:
  hr = '-'*16
  asks = sorted(self.asks, key=lambda e: e.price, reverse=True)
  bids = sorted(self.bids, key=lambda e: e.price, reverse=True)
  return (
    f'{hr}\nASKS\n{hr}\n' +
    '\n'.join(f'{e:{fmt}}' for e in asks) +
    f'\n{hr}\nBIDS\n{hr}\n' +
    '\n'.join(f'{e:{fmt}}' for e in bids) +
    f'\n{hr}'
  )


from trading_sdk import Market
from mexc_sdk import MEXC
from hyperliquid_sdk import Hyperliquid
from dydx_sdk import DYDX

async def dydx_sdk(mnemonic: str | None = None):
  return await DYDX.connect(mnemonic=mnemonic)

async def dydx_market(market: str, *, subaccount: int = 0, mnemonic: str | None = None):
  dydx_client = await dydx_sdk(mnemonic)
  return await dydx_client.market(market, subaccount=subaccount)

def mexc_sdk(api_key: str | None = None, api_secret: str | None = None):
  return MEXC.new(api_key, api_secret)

async def mexc_spot(instrument: str, *, api_key: str | None = None, api_secret: str | None = None):
  mexc = mexc_sdk(api_key, api_secret)
  return await mexc.spot(instrument)

async def mexc_perp(instrument: str, *, api_key: str | None = None, api_secret: str | None = None):
  mexc = mexc_sdk(api_key, api_secret)
  return await mexc.perp(instrument)

def hyperliquid_sdk(address: str | None = None, *, wallet: str | None = None):
  return Hyperliquid.new(address=address, wallet=wallet)

async def hyperliquid_spot(address: str | None = None, *, wallet: str | None = None):
  hyperliquid = hyperliquid_sdk(address, wallet=wallet)
  return await hyperliquid.spot()

async def hyperliquid_spot_market(base: str, quote: str, index: int | None = None, *, address: str | None = None, wallet: str | None = None):
  spot = await hyperliquid_spot(address=address, wallet=wallet)
  market = spot.find(base, quote)
  if index is not None and market.asset_idx != index:
    raise ValueError(f'Expected market {base}/{quote} at index {index}, got {market.base_name}/{market.quote_name}')
  return market

async def hyperliquid_perp(dex: str | None = None, *, address: str | None = None, wallet: str | None = None):
  hyperliquid = hyperliquid_sdk(address, wallet=wallet)
  return await hyperliquid.perp(dex)

async def hyperliquid_perp_market(base: str, dex: str | None = None, *, address: str | None = None, wallet: str | None = None):
  perp = await hyperliquid_perp(dex, address=address, wallet=wallet)
  if dex:
    base = f'{dex}:{base}'
  return perp.find(base)

async def load_market(id: str) -> Market:
  """Load market SDK from ID.
  
  ### Supported exchanges
  - dYdX: `dydx:<subaccount>:<base>-USD`, e.g. `dydx:0:BTC-USD`
  - MEXC Spot: `mexc:spot:<instrument>`, e.g. `mexc:spot:BTCUSDT`
  - MEXC Perp: `mexc:perp:<instrument>`, e.g. `mexc:perp:BTC_USDT`
  - Hyperliquid Spot: `hyperliquid:spot:<base>/<quote>(:<index>)?`, e.g. `hyperliquid:spot:UBTC/USDC:142`
  - Hyperliquid Perp: `hyperliquid:perp:<dex>:<base>`, e.g. `hyperliquid:perp::BTC`
  """
  venue, rest = id.split(':', 1)
  if venue == 'dydx':
    subaccount, market = rest.split(':', 1)
    return await dydx_market(market, subaccount=int(subaccount))
  elif venue == 'mexc':
    kind, instrument = rest.split(':', 1)
    if kind == 'spot':
      return await mexc_spot(instrument)
    elif kind == 'perp':
      return await mexc_perp(instrument)
    else:
      raise ValueError(f'Invalid market kind: {kind} [{id=:}]')
  elif venue == 'hyperliquid':
    kind, rest = rest.split(':', 1)
    if kind == 'spot':
      ticker, index = rest.split(':', 1) if ':' in rest else (rest, None)
      base, quote = ticker.split('/')
      index = int(index) if index is not None else None
      return await hyperliquid_spot_market(base, quote, index)
    elif kind == 'perp':
      dex, base = rest.split(':', 1)
      return await hyperliquid_perp_market(base=base, dex=dex)
    else:
      raise ValueError(f'Invalid market kind: {kind} [{id=:}]')
  else:
    raise ValueError(f'Invalid venue: {venue} [{id=:}]')



class Error(Exception):
  """Base SDK exception."""
  def __str__(self):
    args = self.args[0] if len(self.args) == 1 else ', '.join(map(str, self.args))
    return f'{self.__class__.__name__}({args})'

class NetworkError(Error):
  """Network error: HTTP error, connection timeout, etc."""
  def __str__(self):
    return super().__str__()

class ValidationError(Error):
  """Validation error: invalid response, invalid data, etc."""
  def __str__(self):
    return super().__str__()

class UserError(Error):
  """User error: invalid request, invalid input, etc."""
  def __str__(self):
    return super().__str__()

class AuthError(Error):
  """Authentication error: invalid API key, invalid API secret, etc."""
  def __str__(self):
    return super().__str__()

class ApiError(Error):
  """Unknown error returned by the API."""
  def __str__(self):
    return super().__str__()

class LogicError(Error):
  """Logic error: invalid assumptions, logic, or other bugs on the SDK side."""
  def __str__(self):
    return super().__str__()



from typing_extensions import Sequence, Any
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime

from trading_sdk.core import SDK

class Orders(SDK):
  @dataclass(kw_only=True)
  class Order:
    id: str
    price: Decimal
    qty: Decimal
    """Signed quantity (netagive -> sell, positive -> buy)"""
    filled_qty: Decimal
    """Signed quantity (netagive -> sell, positive -> buy)"""
    active: bool
    """Whether the order is active in the market."""
    details: Any = None

  @SDK.method
  @abstractmethod
  async def query(self, id: str) -> Order:
    """Fetch the state of the order with the given ID."""

  @SDK.method
  @abstractmethod
  async def open(self) -> Sequence[Order]:
    """Fetch your currently open orders."""
    


from typing_extensions import Sequence, Mapping, MutableMapping, Iterable
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from collections import defaultdict

from .types import Flow, Event, Transaction, D, D2

@dataclass
class EventMatcher:
  flows: Sequence[Flow]
  index: Mapping[str, Mapping[str, Mapping[datetime, MutableMapping[int, Flow]]]]
  max_time_window: timedelta = field(default=timedelta(hours=1), kw_only=True)

  @classmethod
  def of(cls, flows: Sequence[Flow], *, max_time_window: timedelta = timedelta(hours=1)):
    idx = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
    for i, f in enumerate(flows):
      idx[f.asset][f.label][f.time][i] = f
    return cls(flows, idx, max_time_window=max_time_window)

  def candidates(self, f_hat: Flow, *, time_window: timedelta | None = None) -> MutableMapping[int, Flow]:
    group = self.index[f_hat.asset][f_hat.label]
    if time_window is None:
      return group[f_hat.time]
    else:
      ts = [t for t in group.keys() if f_hat.time - time_window <= t <= f_hat.time + time_window]
    out: dict[int, Flow] = {}
    for t in ts:
      out.update(group[t])
    return out

  def use(self, key: int):
    f = self.flows[key]
    del self.index[f.asset][f.label][f.time][key]

  def insert(self, key: int):
    f = self.flows[key]
    self.index[f.asset][f.label][f.time][key] = f

  def match(self, flows: Iterable[Flow]) -> list[int] | None:
    """Match the flows to the flows in the matcher. The matched flows are removed from the matcher, preventing double-matching."""
    if (matches := self._match_impl(flows, time_window=None)) is not None:
      return matches
    time_window = timedelta(seconds=1)
    while time_window < self.max_time_window:
      if (matches := self._match_impl(flows, time_window=time_window)) is not None:
        return matches
      time_window *= 2

  def _match_impl(self, flows: Iterable[Flow], *, time_window: timedelta | None) -> list[int] | None:
    out = []
    for f_hat in flows:
      candidates = self.candidates(f_hat, time_window=time_window)
      if len(candidates) == 0:
        for k in out:
          self.insert(k) # restore the used flows
        return None
      elif len(candidates) == 1:
        k = next(iter(candidates.keys()))
      else:
        k = min(candidates.keys(), key=lambda i: abs(f_hat.change - candidates[i].change))
      self.use(k)
      out.append(k)
    return out


def match_transactions(
  flows: Sequence[Flow],
  events: Sequence[Event[D2]],
) -> tuple[list[Transaction[D|D2]], list[Flow]]:
  """Match the flows to the events. Returns `(matched_txs, unmatched_flows)`"""
  matcher = EventMatcher.of(flows)
  used = set[int]()
  transactions: list[Transaction] = []
  
  for e in events:
    if (matches := matcher.match(e.expected_flows)) is None:
      raise ValueError(f'Could not match the flows for the event: {e}')
    used.update(matches)
    transactions.append(Transaction(
      event=e,
      flows=[flows[i] for i in matches] + list(e.fixed_flows)
    ))

  unused = set(range(len(flows))) - used
  unmatched_flows = [flows[i] for i in unused]
  return transactions, unmatched_flows


from typing_extensions import TypeVar, Any, Callable, Protocol, Generic, Self, overload, Sequence
from abc import ABCMeta
from dataclasses import dataclass
import asyncio
import inspect

T = TypeVar('T')
Fn = TypeVar('Fn', bound=Callable)

@dataclass
class Method:
  name: str
  data: Any

  @staticmethod
  def get(fn) -> 'Method | None':
    return getattr(fn, '__sdk_method__', None)

class SDKMeta(ABCMeta):
  def __new__(cls: type, name: str, bases: tuple[type, ...], dct: dict[str, Any]):
    cls = ABCMeta.__new__(cls, name, bases, dct)
    cls.__sdk_methods__ = {}
    cls.__sdk_fields__ = {}
    for base in bases:
      if (methods := getattr(base, '__sdk_methods__', None)) is not None:
        cls.__sdk_methods__.update(methods)
      if (fields := getattr(base, '__sdk_fields__', None)) is not None:
        cls.__sdk_fields__.update(fields.items())
    for k, v in dct.items():
      if (method := Method.get(v)) is not None:
        cls.__sdk_methods__[k] = method
    for k, v in cls.__annotations__.items():
      if getattr(v, '__sdk_fields__', None) is not None:
        cls.__sdk_fields__[k] = v
    return cls


class Mapper(Protocol, Generic[Fn]):
  def __call__(self, fn: Fn, method: Method, path: Sequence[str]) -> Fn:
    ...

class SDK(metaclass=SDKMeta):
  __sdk_methods__: dict[str, Method] = {}
  __sdk_fields__: dict[str, type['SDK']] = {}

  def __sdk_instrument__(self, *mappers: Mapper[Fn], path: tuple[str, ...] = ()) -> Self:
    attrs = {}
    for method, data in self.__sdk_methods__.items():
      fn = getattr(self, method)
      for mapper in mappers:
        fn = mapper(fn, data, path)
      attrs[method] = fn
    for field in self.__sdk_fields__:
      sdk: SDK = getattr(self, field)
      attrs[field] = sdk.__sdk_instrument__(*mappers, path=path + (field,))

    if (fn := getattr(self, '__call__', None)) is not None and (method := Method.get(fn)) is not None:
      for mapper in mappers:
        fn = mapper(fn, method, path)
      attrs['__call__'] = fn

    sdk = self
    class Proxy:
      def __getattr__(self, name: str) -> Any:
        if name in attrs:
          return attrs[name]
        else:
          return getattr(sdk, name)

      def __setattr__(self, name: str, value: Any) -> None:
        setattr(sdk, name, value)

      def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if '__call__' in attrs:
          return attrs['__call__'](*args, **kwargs)
        else:
          return sdk.__call__(*args, **kwargs) # type: ignore

      def __repr__(self) -> str:
        return sdk.__repr__()

    return Proxy() # type: ignore

  @classmethod
  def __sdk_hierarchy__(
    cls, *, field: str | None = None, indent: int = 0, prefix: str = "", last: bool = False
  ) -> str:
    # Prepare current line
    line = ""
    if indent > 0:
      # Show vertical line and branch
      line += prefix + ("└─ " if last else "├─ ")
    else:
      line += ""
    if field is not None:
      line += f"{field}: "
    line += cls.__name__ + "\n"

    # Prepare new prefix for children
    if indent > 0:
      child_prefix = prefix + ("   " if last else "│  ")
    else:
      child_prefix = ""

    # Print methods
    method_items = list(cls.__sdk_methods__.items())
    field_items = list(cls.__sdk_fields__.items())

    for i, (method, data) in enumerate(method_items):
      is_last = (i == len(method_items) - 1 and not field_items)
      mid_prefix = child_prefix + ("└─ " if is_last else "├─ ")
      line += mid_prefix + method + "()" + "\n"

    # Print fields
    for j, (next_field, sdk) in enumerate(field_items):
      is_last_field = (j == len(field_items) - 1)
      line += sdk.__sdk_hierarchy__(
        indent=indent + 1,
        field=next_field,
        prefix=child_prefix,
        last=is_last_field,
      )

    return line

  @overload
  @classmethod
  def method(cls, fn: Fn, /) -> Fn:
    ...
  @overload
  @classmethod
  def method(cls, /, *, name: str | None = None, data: Any = None) -> Callable[[Fn], Fn]:
    ...
  @classmethod
  def method(cls, fn = None, /, *, name: str | None = None, data: Any = None):
    def decorator(fn: Fn) -> Fn:
      setattr(fn, '__sdk_method__', Method(name=name or fn.__name__, data=data))
      return fn

    if fn is None:
      return decorator
    else:
      return decorator(fn)

S = TypeVar('S', bound=SDK)
def instrument(sdk: S, *mappers: Mapper[Fn]) -> S:
  return sdk.__sdk_instrument__(*mappers)

E = TypeVar('E', bound=Exception, contravariant=True)

class RetryLogger(Protocol, Generic[E]):
  def __call__(
    self, fn: Callable, method: Method, path: Sequence[str], *,
    args: tuple, kwargs: dict, exception: E, retries: int, delay: float
  ):
    ...

def default_retry_logger(
  fn: Callable, method: Method, path: Sequence[str], *,
  args: tuple, kwargs: dict, exception: Exception, retries: int, delay: float
):
  name = getattr(fn, '__name__', method.name)
  path_str = '.'.join(path)
  print(f'Exponential retry [{retries=}, {delay=:.2f}s]. Calling {path_str}.{name} with {args=}, {kwargs=}. Exception:', exception)

def exponential_retry(
  *exceptions: type[E],
  max_retries: int | None = None, base_delay: float = 1.0,
  max_delay: float | None = None, log: RetryLogger[E] | None = default_retry_logger
) -> Mapper:
    def exponential_retry_wrapper(fn: Fn, method: Method, path: Sequence[str]) -> Fn:
      if inspect.iscoroutinefunction(fn):
        async def exponential_retry_wrapped(*args, **kwargs):
          retries = 0
          while True:
            try:
              return await fn(*args, **kwargs)
            except exceptions as e:
              if max_retries is not None and retries >= max_retries:
                raise e
              retries += 1
              delay = base_delay * 2**retries
              if max_delay is not None and delay > max_delay:
                delay = max_delay
              if log is not None:
                log(fn, method, path, args=args, kwargs=kwargs, exception=e, retries=retries, delay=delay)
              await asyncio.sleep(delay)
        return exponential_retry_wrapped # type: ignore

      else:
        return fn

    return exponential_retry_wrapper

class Logger(Protocol):
  def __call__(
    self, fn: Callable, method: Method, path: Sequence[str], *,
    args: tuple, kwargs: dict
  ):
    ...

def default_logger(
  fn: Callable, method: Method, path: Sequence[str], *,
  args: tuple, kwargs: dict
):
  path_str = '.'.join(path)
  name = getattr(fn, '__name__', method.name)
  print(f'Calling {path_str}.{name} with {args=}, {kwargs=}')

def log(logger: Logger = default_logger) -> Mapper:
  def log_wrapper(fn: Fn, method: Method, path: Sequence[str]) -> Fn:
    if inspect.iscoroutinefunction(fn):
      async def coro_log_wrapped(*args, **kwargs):
        logger(fn, method, path, args=args, kwargs=kwargs)
        return await fn(*args, **kwargs)
      return coro_log_wrapped # type: ignore

    elif inspect.isasyncgenfunction(fn):
      async def agen_log_wrapped(*args, **kwargs):
        logger(fn, method, path, args=args, kwargs=kwargs)
        async for x in fn(*args, **kwargs):
          yield x
      return agen_log_wrapped # type: ignore

    elif inspect.isfunction(fn):
      def log_wrapped(*args, **kwargs):
        logger(fn, method, path, args=args, kwargs=kwargs)
        return fn(*args, **kwargs)
      return log_wrapped # type: ignore

    else:
      return fn

  return log_wrapper



from typing_extensions import Sequence, Any
from abc import abstractmethod
from dataclasses import dataclass
import asyncio

from trading_sdk.core import SDK

class Cancel(SDK):

	@dataclass(kw_only=True)
	class Result:
		details: Any = None

	@SDK.method
	@abstractmethod
	async def order(self, id: str) -> Result:
		"""Cancel an order."""
		
	@SDK.method
	async def orders(self, ids: Sequence[str]) -> Any:
		"""Cancel multiple orders."""
		return await asyncio.gather(*[self.order(id) for id in ids])

	@SDK.method
	@abstractmethod
	async def open(self) -> Any:
		"""Cancel all open orders."""



from typing_extensions import Any
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal

from trading_sdk.core import SDK, trunc2tick, round2tick, ceil2tick

class Rules(SDK):
  @dataclass(kw_only=True)
  class Rules:
    """Market rules type."""
    base: str
    """Base asset of the instrument."""
    quote: str
    """Quote asset of the instrument."""
    fee_asset: str
    """Asset used for fees/funding payments."""
    tick_size: Decimal
    """Tick size of the price (in quote units)."""
    step_size: Decimal
    """Step size of the quantity (in base units)."""
    fixed_min_qty: Decimal | None = None
    """Minimum quantity of the order (in base units)."""
    min_value: Decimal | None = None
    """Minimum value of the order (in quote units)."""
    max_qty: Decimal | None = None
    """Maximum quantity of the order (in base units)."""
    fixed_min_price: Decimal | None = None
    """Minimum price of the order (in quote units)."""
    rel_min_price: Decimal | None = None
    """Minimum price of the order (in quote units), relative to the current price (e.g. 0.95 = 5% below the current price)."""
    rel_max_price: Decimal | None = None
    """Maximum price of the order (in quote units), relative to the current price (e.g. 1.05 = 5% above the current price)."""
    fixed_max_price: Decimal | None = None
    """Maximum price of the order (in quote units)."""
    maker_fee: Decimal
    """Maker fee of the order (in quote units)."""
    taker_fee: Decimal
    """Taker fee of the order (in quote units)."""
    api: bool
    """Whether the instrument can be traded via API."""
    details: Any = None
    """Raw details of the rules."""

    def min_price(self, mark_price: Decimal, /) -> Decimal | None:
      """Minimum price of the order (in quote units), accounting for the minimum value and fixed minimum price."""
      values: list[Decimal] = []
      if self.fixed_min_price is not None:
        values.append(self.fixed_min_price)
      if self.rel_min_price is not None:
        values.append(self.rel_min_price * mark_price)
      return min(values, default=None)

    def max_price(self, mark_price: Decimal, /) -> Decimal | None:
      """Maximum price of the order (in quote units), accounting for the maximum value and fixed maximum price."""
      values: list[Decimal] = []
      if self.fixed_max_price is not None:
        values.append(self.fixed_max_price)
      if self.rel_max_price is not None:
        values.append(self.rel_max_price * mark_price)
      return max(values, default=None)

    def min_qty(self, price: Decimal) -> Decimal:
      """Minimum quantity of the order (in base units), accounting for the minimum value and fixed minimum quantity."""
      min_qty = self.fixed_min_qty or self.step_size
      if self.min_value is not None:
        min_qty = max(min_qty, self.min_value / price)
      return ceil2tick(min_qty, self.step_size)
    
    def trunc_qty(self, base_qty: Decimal, *, price: Decimal) -> Decimal | None:
      """Truncate the (base asset) quantity to the nearest step size. Returns `None` if the quantity is too small."""
      qty = trunc2tick(base_qty, self.step_size)
      if qty > self.min_qty(price):
        return qty

    def round_price(self, price: Decimal) -> Decimal:
      """Round the price to the nearest tick size."""
      return round2tick(price, self.tick_size)
    
    def amount2qty(self, quote_amount: Decimal, *, price: Decimal) -> Decimal | None:
      """Convert a quote amount to a base quantity, truncating to the nearest step size. Returns `None` if the quantity is too small."""
      return self.trunc_qty(quote_amount / price, price=price)
    
    def qty2amount(self, base_qty: Decimal, *, price: Decimal) -> Decimal:
      """Convert a base quantity to a quote amount."""
      return base_qty * price
  
  @SDK.method
  @abstractmethod
  async def get(self) -> Rules:
    """Fetch market rules."""

  @SDK.method
  async def __call__(self) -> Rules:
    """Fetch market rules."""
    return await self.get()